#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"
#include "serialfunc.h"

/* L, Lg, M, Mg, and others */
double quadratic(const int dim, int *x, double **a) {
    double ret = 0;
    int i, j;

    for (i=0; i<dim; i++) {
        double tmp = 0;
        for (j=0; j<dim; j++) {
            tmp += a[i][j] * x[j];
        }
        ret += tmp * x[i];
    }
    return ret/2;
}

double quadratic2(const int dim, int *x, double **a, double *b) {
    double ret = 0;
    int i;

    for (i=0; i<dim; i++) ret += x[i] * b[i];
    return quadratic(dim, x, a) + ret;
}

double quadratic3(const int dim, int *x, double **a, double *b, double c) {
    return quadratic2(dim, x, a, b) + c;
}

/* easy book p.41 (4.21) */
int is_lgconv_quadratic(const int dim, double **a) {
    int i, j;
    double sum;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
            if (fcmp(a[i][j],0) > 0)        return FALSE;
        }
        sum = 0;
        for (j=0; j<dim; j++) sum += a[i][j];
        if (fcmp(sum,0) < 0) return FALSE;
    }
    return TRUE;
}

/* easy book p.42 (4.23) */
int is_lconv_quadratic(const int dim, double **a) {
    int i, j;
    double sum;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
            if (fcmp(a[i][j],0) > 0)        return FALSE;
        }
        sum = 0;
        for (j=0; j<dim; j++) sum += a[i][j];
        if (fcmp(sum,0) != 0) return FALSE;
    }
    return TRUE;
}

/* easy book p.65 (5.15)(5.16) */
int is_mgconv_quadratic(const int dim, double **a) {
    int i, j, k;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = a[i][j];
            if (fcmp(tmp,a[j][i]) != 0) return FALSE;
            if (fcmp(tmp,0) < 0)        return FALSE;
            for (k=0; k<dim; k++) {
                if (k != i && k != j &&
                    fcmp(tmp,a[i][k]) < 0 &&
                    fcmp(tmp,a[j][k]) < 0) return FALSE;
            }
        }
    }
    return TRUE;
}

/* easy book p.68 (5.27) */
int is_mconv_quadratic(const int dim, double **a) {
    int i, j, k, l;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
    for (k=0; k<dim; k++) {
        if (k==i || k==j) continue;
        for (l=0; l<=k; l++) {
            if (l==i || l==j) continue;
            if (fcmp(a[i][j]+a[k][l],a[i][k]+a[j][l]) < 0 &&
                fcmp(a[i][j]+a[k][l],a[i][l]+a[j][k]) < 0) return FALSE;
        }
    }                    
        }
    }                    
    return TRUE;
}

/* Lg = L */
/* easy book p.42 (4.26) */
void make_hesse_lg(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  oracle_call++;  x[i]++;
            tmp += f(dim,x);  oracle_call++;  x[j]++;
            tmp -= f(dim,x);  oracle_call++;  x[i]--;
            tmp += f(dim,x);  oracle_call++;  x[j]--;
            a[i][j] = a[j][i] = -tmp;
        }
    }
    /* diagonal elements */
    for (i=0; i<dim; i++) {
        double tmp = 0;
        tmp += f(dim,x);  oracle_call++;  x[i]++;
        tmp -= f(dim,x);  oracle_call++;  for (j=0; j<dim; j++) x[j]++;
        tmp += f(dim,x);  oracle_call++;  x[i]--;
        tmp -= f(dim,x);  oracle_call++;  for (j=0; j<dim; j++) x[j]--;
        for (j=0; j<dim; j++) if (i != j) tmp -= a[i][j];
        a[i][i] = tmp;
    }
}


/* Mg */
/* easy book p.69 (5.28) */
void make_hesse_mg(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  oracle_call++;  x[i]++;
            tmp += f(dim,x);  oracle_call++;  x[j]++;
            tmp -= f(dim,x);  oracle_call++;  x[i]--;
            tmp += f(dim,x);  oracle_call++;  x[j]--;
            a[i][j] = a[j][i] = -tmp;
        }
    }
}
 
/* M */
/* H. Hirai and K. Murota:
   M-convex functions and tree metrics,
   Japan Journal of Industrial and Applied Mathematics,
   21 (2004), 391--403.
*/
void make_hesse_m(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;
    double f0 = f(dim, x);

    for (i=0; i<dim; i++) {
	a[i][i] = 0;  /* diagonal elements */
        for (j=0; j<i; j++) {
            double tmp = f0;
            x[i]++;  x[j]--;  tmp -= f(dim,x)/2; oracle_call++;
            x[i]-=2; x[j]+=2; tmp -= f(dim,x)/2; oracle_call++;
            x[i]++;  x[j]--;
            a[i][j] = a[j][i] = tmp;
        }
    }
}


/* M */
/* not written in any article */
/* attention: Hesse matrix is of Mg and its size is (dim-1)*(dim-1) */
void make_hesse_m2(const int dim, double f(int,int*), int *x, double **a) {
    int i, j;

    for (i=0; i<dim-1; i++) {
        for (j=0; j<=i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  oracle_call++;  x[i]++;  x[dim-1]--;
            tmp += f(dim,x);  oracle_call++;  x[j]++;  x[dim-1]--;
            tmp -= f(dim,x);  oracle_call++;  x[i]--;  x[dim-1]++;
            tmp += f(dim,x);  oracle_call++;  x[j]--;  x[dim-1]++;
            a[i][j] = a[j][i] = -tmp;
        }
    }
}



void make_liner_const(const int dim, double f(int,int*), int *x,
		      double **a, double *b, double *c) {
    int i, step = 10;
    double tmp;

    tmp = f(dim,x) - quadratic(dim,x,a);  oracle_call++;
    for (i=0; i<dim; i++) {
        x[i] += step;
        b[i] = (f(dim,x) - quadratic(dim,x,a) - tmp) / step;  oracle_call++;
        x[i] -= step;
    }

    *c = f(dim,x) - quadratic2(dim,x,a,b);  oracle_call++;
}


void make_liner_const_m(const int dim, double f(int,int*), int *x,
                        double **a, double *b, double *c) {
    int i, step = 10;
    double tmp;

    tmp = f(dim,x) - quadratic(dim,x,a);  oracle_call++;
    b[dim-1] = 0;
    x[dim-1] -= step;
    for (i=0; i<dim-1; i++) {
        x[i] += step;
        b[i] = (f(dim,x) - quadratic(dim,x,a) - tmp) / step;  oracle_call++;
        x[i] -= step;
    }
    x[dim-1] += step;

    *c = f(dim,x) - quadratic2(dim,x,a,b);  oracle_call++;
}


#ifdef TEST

double f_Lg(int dim, int *x) {
    double a1[3] = {  8, -3, -1 };
    double a2[3] = { -3,  5, -2 };
    double a3[3] = { -1, -2,  5 };
    double *a[3] = { a1, a2, a3 };
    double  b[3] = { -1, -1, -1 };
    assert(dim == 3);
    return quadratic2(3, x, a, b);
}

double f_L(int dim, int *x) {
    double a1[3] = {  4, -3, -1 };
    double a2[3] = { -3,  5, -2 };
    double a3[3] = { -1, -2,  3 };
    double *a[3] = { a1, a2, a3 };
    double  b[3] = { -1, -1, -1 };
    assert(dim == 3);
    return quadratic2(3, x, a, b);
}

double f_Mg(int dim, int *x) {
    double a1[3] = {  1,  1,  1 };
    double a2[3] = {  1,  2,  2 };
    double a3[3] = {  1,  2,  3 };
    double *a[3] = { a1, a2, a3 };
    double  b[3] = { -1, -1, -1 };
    assert(dim == 3);
    return quadratic2(3, x, a, b);
}

double f_none(int dim, int *x) {
    double a1[3] = {  -100,  -1,  1 };
    double a2[3] = {  -1,  2,  2 };
    double a3[3] = {  1,  2,  3 };
    double *a[3] = { a1, a2, a3 };
    double  b[3] = { -1, -1, -1 };
    assert(dim == 3);
    return quadratic2(3, x, a, b);
}

/* L#-conv */
double fsep(int dim, int *x) {
    int v[] = {
        'L', 1, 2, 1000, /* 'L' needs three arguments */
        0, 1,         /* two indexes of x[], no end mark */
        '2', 2, 3, 4, /* '2' needs three arguments */
        2, 3,        /* one index of x[], -1 means ommited */
        '0',          /* end of equation */
    };
    int i; double y[4];
    assert(dim == 4);
    for (i=0; i<dim; i++) y[i] = x[i];
    return fpseparable(dim, y, v);
}

/* M#-conv */
double flam(int dim, int *x) {
    int v[] = {
        'E', 2, 3, 4, /* 'E' needs three arguments */
        0, 1, 2, -1,  /* indexes of x[], -1 means end */
        '0',          /* end of equation */
    };
    int i; double y[4];
    assert(dim == 4);
    for (i=0; i<dim; i++) y[i] = x[i];
    return flaminar(dim, y, v);
}

void test_v(int dim, double **a) {
    fmatrix_print(dim, a);
    printf("\n");
    if (is_lgconv_quadratic(dim, a)) printf("L#-conv func\n");
    if ( is_lconv_quadratic(dim, a)) printf("L -conv func\n");
    if (is_mgconv_quadratic(dim, a)) printf("M#-conv func\n");
    if ( is_mconv_quadratic(dim, a)) printf("M -conv func\n");
    printf("\n");
}

void test_i(int dim, double f(int,int*), double **a, int *x, int *y) {
    double  *b = fvector_new(dim);
    double   c;

    make_liner_const(dim, f, x, a, b, &c);
    printf("%g, %g :", f(dim, x), quadratic3(dim, x, a, b, c));
    printf("%g, %g\n", f(dim, y), quadratic3(dim, y, a, b, c));
}

int main(void) {
    double **a = fmatrix_new(4, 4);
    int     x[3] = {  1,  3, -4 };
    int     y[3] = { -1, -2,  3 };
    int i;

    make_hesse_lg(3, f_Lg, x, a); fmatrix_print(3, a); printf("\n");
    make_hesse_mg(3, f_Lg, x, a); fmatrix_print(3, a); printf("\n");
    test_i(3, f_Lg, a, x, y);
    test_v(3, a);

    for (i=0; i<3; i++) x[i]-=2;
    make_hesse_lg(3, f_L, x, a); fmatrix_print(3, a); printf("\n");
    make_hesse_mg(3, f_L, x, a); fmatrix_print(3, a); printf("\n");
    test_i(3, f_L, a, x, y);
    for (i=0; i<3; i++) x[i]++;
    make_hesse_lg(3, f_L, x, a);
    test_i(3, f_L, a, x, y);
    for (i=0; i<3; i++) x[i]++;
    make_hesse_lg(3, f_L, x, a);
    test_i(3, f_L, a, x, y);
    test_v(3, a);

    make_hesse_lg(3, f_Mg, x, a); fmatrix_print(3, a); printf("\n");
    make_hesse_mg(3, f_Mg, x, a); fmatrix_print(3, a); printf("\n");
    test_i(3, f_Mg, a, x, y);
    test_v(3, a);

    make_hesse_m(3, f_Mg, x, a);
    test_i(3, f_Mg, a, x, y);
    test_v(3, a);

    make_hesse_m2(3, f_Mg, x, a);
    printf("%g, %g\n", f_Mg(3, x), quadratic(3, x, a));
    test_v(2, a);

    make_hesse_mg(3, f_none, x, a);
    test_i(3, f_none, a, x, y);
    test_v(3, a);

    make_hesse_mg(4, fsep, x, a); fmatrix_print(4, a); printf("\n");
    make_hesse_lg(4, fsep, x, a); fmatrix_print(4, a); printf("\n");
    test_i(4, fsep, a, x, y);
    test_v(4, a);

    make_hesse_lg(4, flam, x, a); fmatrix_print(4, a); printf("\n");
    make_hesse_mg(4, flam, x, a); fmatrix_print(4, a); printf("\n");
    test_i(4, flam, a, x, y);
    test_v(4, a);

    return 0;
}

/* usage:
   gcc -DTEST quadratic.c util.o cpu_time.o SFMT.o serialfunc.o && ./a.out
*/
#endif
